// dp_bcc.cc -- find CC of max TF path integral with dynamic programming

// To use this file, your version of Octave must support dynamic
// linking.  To find out if it does, type the command
//
//   x = octave_config_info; x.DEFS
//
// at the Octave prompt.  Support for dynamic linking is included if
// the output contains the string -DWITH_DYNAMIC_LINKING=1.
//
// To compile this file, type the command
//
// mkoctfile --verbose dp_bcc.cc
//
// at the shell prompt.  The script mkoctfile should have been
// installed along with Octave.  Running it will create a file called
// dp_bcc.oct that can be loaded by Octave.
//
// Enter:
//
// help dp_bcc
//
// at the Octave prompt to have a description of how dp_bcc should be used.  
// 
// This file may be used under the terms defined by the GNU 
// General Public License [write to the Free Software Foundation, 
// 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA]
//
// WITHOUT ANY WARRANTY; without even the implied warranty 
// of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
//
// Copyright (C) 2005 Eric Chassande-Mottin, CNRS (France)

//

#include <octave/config.h>

#include <iostream>
#include <sstream>
using namespace std;
#include <string>

#include <octave/defun-dld.h>
#include <octave/error.h>
#include <octave/oct-obj.h>
#include <octave/pager.h>
#include <octave/symtab.h>
#include <octave/variables.h>

static int Nr1,step,Nr2,debug_flag;

static inline int
getmin ( int x, int y)
{
  return x < y ? x : y;
}

static void
first_interval(ColumnVector& bak, const Matrix& buf, const int* idx)
{

  double s,a,y,dy;
  int i,k,l,m,n,ny;

  int b=buf.columns();
  int r=buf.rows();

  for (i=0 ; i<r ; i++)
    {
      m=idx[i];

      /* 1st case: const frequency */
      s=0.0;
      for (n=0 ; n<b ; n++)
	s+=buf(i,n);
      bak(m)=s;

      /* 2nd case: chirplet with slope > 0 */
      l=1; k=i-step;
      while ((l<=Nr1)&(k>=0))
	{
	  /* compute line integral */
	  a=((double) (i-k))/((double) b);
	  s=buf(k,0);
	  for (int n=1 ; n<b ; n++)
	    {
	      y=k+a*n;
	      ny=(int) y; 
	      dy=y-ny;
	      s+=(1-dy)*buf(ny,n)+dy*buf(ny+1,n);
	    }
	  bak(m-l)=s;

	  k-=step; l++;
	}

      /* 3rd case: chirplet with slope < 0 */
      l=1; k=i+step;
      while ((l<=Nr1)&(k<r))
	{
	  /* compute line integral */
	  a=((double) (i-k))/((double) b);
	  s=buf(k,0);
	  for (int n=1 ; n<b ; n++)
	    {
	      y=k+a*n;
	      ny=(int) y; 
	      dy=y-ny;
	      s+=(1-dy)*buf(ny,n)+dy*buf(ny+1,n);
	    }
	  bak(m+l)=s;

	  k+=step; l++;
	}
    }
}

static void
next_interval(ColumnVector& tot, const ColumnVector& bak, int* rec, const Matrix& buf, const int* idx)
{

  double s,t,a,y,dy;
  int i,k,l,m,n,ny,km,ll,kk,mm;

  int b=buf.columns();
  int r=buf.rows();

  for (i=0 ; i<r ; i++)
    {
      m=idx[i];

      /* 1st case: const frequency */
      s=0.0;
      for (n=0 ; n<b ; n++)
	s+=buf(i,n);
 
      /* 1st case (1): delta^1=0, delta^2=0 */
      tot(m)=bak(m)+s;
      rec[m]=m;
      
      /* 1st case (2): delta^1=0, delta^2<0 */
      l=1; k=i-step;
      while ((k>=0)&(l<=Nr2)&(l<=Nr1))
	{
	  t=bak(m-l)+s;
	  if (t>tot(m))
	    {
	      tot(m)=t;
	      rec[m]=m-l;
	    }
	  k-=step; l++; 
	}
      
      /* 1st case (3): delta^1=0, delta^2>0 */
      l=1; k=i+step;
      while ((k<r)&(l<=Nr2)&(l<=Nr1))
	{
	  t=bak(m+l)+s;
	  if (t>tot(m))
	    {
	      tot(m)=t;
	      rec[m]=m+l;
	    }
	  k+=step;l++;
	}

      /* 2nd case: delta^1 > 0 */
      l=1; k=i-step;
      while ((k>=0)&(l<=Nr1))
	{
	  /* compute line integral */
	  a=((double) (i-k))/((double) b);
	  s=buf(k,0);
	  for (int n=1 ; n<b ; n++)
	    {
	      y=k+a*n;
	      ny=(int) y; 
	      dy=y-ny;
	      s+=(1-dy)*buf(ny,n)+dy*buf(ny+1,n);
	    }
	  
	  // bin where starts beam which gives delta^2=0 is: k-l*step
	  km=k-l*step;	      
	  if (km<0) // (1) start bin is outside matrix
	    {
	      // are there any chirplet giving |delta^2|<max and starting in the matrix?
	      ll=getmin(Nr2,l+Nr1); 
	      kk=km+ll*step; // bin where starts top admissible chirplet
	      if (kk<0) // (2) start bin in outside the matrix
		{
		  mm=idx[k];     // horizontal chirplet ending in k
		  tot(m-l)=bak(mm)+s; // choose horizontal beam ending in k
		  rec[m-l]=mm;		  
		}
	      else // (2) start bin in inside the matrix
		{
		  /* init */
		  mm=idx[k]+(ll-l);
		  tot(m-l)=bak(mm)+s;
		  rec[m-l]=mm;
		  
		  /* scan down */
		  ll=1; kk-=step;
		  while (kk>=0)
		    {
		      t=bak(mm-ll)+s;
		      if (t>tot(m-l))
			{
			  tot(m-l)=t;
			  rec[m-l]=mm-ll;
			}
		      kk-=step; ll++;
		    }
		}
	    }
	  else
	    {
	      /* 2nd case (1): delta^1>0, delta^2=0 */
	      mm=idx[k]-l;
	      tot(m-l)=bak(mm)+s;
	      rec[m-l]=mm;
	      
	      /* 2nd case (2): delta^1>0, delta^2<0 */
	      ll=1; kk=km-step;
	      while ((kk>=0)&(ll<=Nr2)&(ll+l<=Nr1))
		{
		  t=bak(mm-ll)+s;
		  if (t>tot(m-l))
		    {
		      tot(m-l)=t;
		      rec[m-l]=mm-ll;
		    }
		  kk-=step; ll++;
		}
	      
	      /* 2nd case (3): delta^1>0, delta^2>0 */
	      ll=1; kk=km+step;
	      while ((kk<r)&(ll<=Nr2)&(ll-l<=Nr1))
		{
		  t=bak(mm+ll)+s;
		  if (t>tot(m-l))
		    {
		      tot(m-l)=t;
		      rec[m-l]=mm+ll;
		    }
		  kk+=step; ll++;
		}
	    }
	  k-=step;l++;
	}
      
      /* 3rd case: delta^1 < 0 */
      l=1; k=i+step;
      while ((k<r)&(l<=Nr1))
	{
	  /* compute line integral */
	  a=((double) (i-k))/((double) b);
	  s=buf(k,0);
	  for (int n=1 ; n<b ; n++)
	    {
	      y=k+a*n;
	      ny=(int) y; 
	      dy=y-ny;
	      s+=(1-dy)*buf(ny,n)+dy*buf(ny+1,n);
	    }
	  
	  // bin where starts beam which gives delta^2=0 is: k+l*step
	  km=k+l*step;
	  if (km>=r) // (1) start bin is outside matrix
	    {
	      // are there any chirplet giving |delta^2|<max and starting in the matrix?
	      ll=getmin(Nr2,l+Nr1); 
	      kk=km-ll*step; // select bottom admissible beam
	      if (kk>=r) // (2) start bin in outside the matrix
		{
		  mm=idx[k];     // horizontal chirplet ending in k
		  tot(m+l)=bak(mm)+s; // choose horizontal beam by default
		  rec[m+l]=mm;	// choose horizontal chirplet ending in k
		}
	      else // (2) start bin in inside the matrix
		{
		  /* init */
		  mm=idx[k]-(ll-l);
		  tot(m+l)=bak(mm)+s;
		  rec[m+l]=mm;
		  
		  /* scan up */
		  ll=1; kk+=step;
		  while (kk<r)
		    {
		      t=bak(mm+ll)+s;
		      if (t>tot(m+l))
			{
			  tot(m+l)=t;
			  rec[m+l]=mm+ll;
			}
		      kk+=step; ll++;
		    }
		}
	    }
	  else
	    {
	      /* 3rd case (1): delta^1<0, delta^2=0 */
	      mm=idx[k]+l;
	      tot(m+l)=bak(mm)+s;
	      rec[m+l]=mm;
	      
	      /* 3rd case (2): delta^1<0, delta^2<0 */
	      ll=1; kk=km-step;
	      while ((kk>=0)&(ll<=Nr2)&(ll-l<=Nr1))
		{
		  t=bak(mm-ll)+s;
		  if (t>tot(m+l))
		    {
		      tot(m+l)=t;
		      rec[m+l]=mm-ll;
		    }
		  kk-=step; ll++;
		}
	      
	      /* 2nd case (3): delta^1>0, delta^2>0 */
	      ll=1; kk=km+step;
	      while ((kk<r)&(ll<=Nr2)&(ll+l<=Nr1))
		{
		  t=bak(mm+ll)+s;
		  if (t>tot(m+l))
		    {
		      tot(m+l)=t;
		      rec[m+l]=mm+ll;
		    }
		  kk+=step; ll++;
		  
		}
	    }
	  k+=step;l++;
	}
    }
}

DEFUN_DLD (dp_bcc, args, ,
  "DP_BCC:  find CC of max path integral with dynamic programming\n")
{
  octave_value_list retval;
  
  /*----------- get input parameters ----------------*/
  int nargin = args.length ();
  
  if (nargin < 5)
    {
      // print_usage("dp_bcc");
      error("Usage: out = dp_bcc(in,b,Nr1,step,Nr2[,debug])");
      return retval;
    }

  /* get the data */
  Matrix in=args(0).matrix_value();
  int c=in.columns();
  int r=in.rows();

  /* check that c is a power of 2 */
  while(!(c & 1))
      c = c>>1;

  if (!(c==1))
    {
      error("dp_bcc: in[row,col] ; col should be a power of 2");
      return retval;
    }
  c=in.columns();

  /* get the size of the interval */
  int b =args(1).int_value();

  /* check that b is a power of 2 */
  while(!(b & 1))
      b = b>>1;

  if (!(b==1))
    {
      error("dp_bcc: b should be a power of 2");
      return retval;
    }
  b=args(1).int_value();

  if ((b<2)|(b>=c))
    {
      error("dp_bcc: b should be 0 < b < columns(in)");
      return retval;
    }
  int Nt=(int) c/b;

  /* get the number of steps, constraint 1: 1st derivative */
  Nr1 =args(2).int_value();

  /* get the step size*/
  step =args(3).int_value();

  /* get the number of steps, constraint 2: 2nd derivative */
  Nr2 =args(4).int_value();
  if (Nr2>=(2*Nr1))
      printf("dp_bcc: warning! no effective constraint on 2nd derivative");

  /* get debug flag */
  debug_flag=0;
  if (nargin==6)
    debug_flag=args(5).int_value();
  
  /* paths counting and numbering */
  int* h_idx=(int*) malloc(r*sizeof(int)); // record index of horiz chirplet
  int* u_idx=(int*) malloc(r*sizeof(int)); // record index of max neg slope chirplet

  int n=0,k,l;
  for (int i=0; i<r; i++)
    {
      l=0; k=i-step;
      while ((k>=0)&(l<Nr1))
	{
	  k-=step;
	  l++; n++;
	}
      h_idx[i]=n; n++;
      l=0; k=i+step;
      while ((k<r)&(l<Nr1))
	{
	  k+=step;
	  l++; n++;
	}
      u_idx[i]=n-1;
      // printf("h_idx[%d]=%d, u_idx[%d]=%d\n",i,h_idx[i],i,u_idx[i]);
    }
  int Nc=n;  
  if (debug_flag==1)
    printf("Nc=%d\n",Nc);

  /* current and backup states */
  ColumnVector tot(Nc);
  tot.fill(0.0);
  ColumnVector bak(Nc);
  bak.fill(0.0);

  /* init path table: rec[n,m] is the row of the previous site (m-1) */
  int** rec=(int**) malloc((Nt-1)*sizeof(int*)); // alloc rows
  for (int j=0 ; j<(Nt-1) ; j++)
    rec[j]=(int*) malloc(Nc*sizeof(int)); //alloc columns 

  for(int j=0;j<(Nt-1);j++)
    for(int i=0;i<Nc;i++)
      *(rec[j]+i)=-1;

  Matrix interval(r,b,0.0);
  interval=in.extract(0,0,r-1,b-1);
  first_interval(bak,interval,h_idx);
  
  if (debug_flag==1)
    {
      printf("init: \n");
      int jj=r-1;
      for (int ii=Nc-1 ; ii>-1 ; ii--)
	{
	  if (ii==h_idx[jj])
	    {	
	      printf("* %d- ",jj); jj--;
	    }
	  printf("%d: bak=%2.2f\n",ii,bak(ii));
	}
      printf("\n\n");	
    }

  /* main loop */
  int jcol=b;
  for (int j=0 ; j<(Nt-1) ; j++)
    {
      interval=in.extract(0,jcol,r-1,jcol+b-1);
      next_interval(tot,bak,rec[j],interval,h_idx);
      bak=tot;
      jcol+=b;

      if (debug_flag==1)
	{
	  printf("\ninterval %d\n",j);
	  int kk=r-1;
	  for (int ii=Nc-1 ; ii>-1 ; ii--)
	    {
	      if (ii==h_idx[kk])
		{
		  printf("row %d chl %d:",kk,ii); kk--;
		}
	      else
		  printf("       chl %d:",ii);
	      for (int jj=0 ; jj<(Nt-1) ; jj++)
		printf("%d ",*(rec[jj]+ii));
	      printf(" = %2.2f\n",tot(ii));
	    }
	  printf("\n\n");
	}
    }

  /* compute stat and get max */
  n=0;
  double totmax=tot(0);
  for (int i=1; i<Nc ; i++)
    {
      if (tot(i)>totmax)
	{
	  totmax=tot(i);
	  n=i;
	}
    }

  /* trace chain back */
  ColumnVector index(Nt+1);
  k=0;
  while (u_idx[k]<n)
    k++;
  index(Nt)=k+1;
  if (debug_flag==1)
    printf("\ntrace back:\n %d <-",n);
  
  for (int j=Nt-2;j>-1;j--)
    {
      n=*(rec[j]+n);
      k=0;
      while (u_idx[k]<n)
	k++;
      index(j+1)=k+1;
      if (debug_flag==1)
	printf(" %d <-",n);
    }
  index(0)=index(1)+step*(n-h_idx[k]);
  if (debug_flag==1)
    printf("\n\n");

  // check connection of all chains  
  if (debug_flag==2)
    {
      int prev, next;
      for (int i=0; i<Nc ; i ++)
	{
	  k=0;
	  while (u_idx[k]<i)
	    k++;	  
	  prev=i;
	  printf("row %d: %d <- ",k,i);
	  for (int j=Nt-2;j>-1;j--)
	    {
	      next=*(rec[j]+prev);
	      prev=next;
	      printf("%d <- ",prev);
	    }
	  printf("\n");
	}
    }

  /* free mem */
  free(h_idx);
  h_idx=NULL;
  free(u_idx);
  u_idx=NULL;
  for(int j=0;j<(Nt-1);j++)
      free(rec[j]);
  free(rec);
  rec=NULL;

  retval(0)=totmax;
  retval(1)=index;

  return retval;
}
