%%
%% Copyright (C) 2005 Eric Chassande-Mottin, CNRS (France)
%%
%% This program may be freely copied under the terms of the 
%% GNU General Public License (GPL), version 2, or at your 
%% option any later version.
%%

function [x,ifl,difl,ddifl,chirp_mass]=gwchirp(N,f_in,t_out,fs,phi_star)
  %% generates a typical GW chirp like signal (no amplitude modulation).
  %% Usage:       [x,ifl,difl,ddifl,chirp_mass]=gwchirp(N,[f_in,t_out,fs,phi_star])
  %%
  %% input parameters:    N = signal size
  %%                      f_in = input (at t=0) reduced frequency  
  %%                      [default= .1, should be 0 < f_in < 1/2]
  %%                      t_out = time index at which the signal freq reaches 1/2
  %%                      [default= N, should be 1 <= t_out <= N]
  %%                      fs = sampling frequency (used to calibrate the
  %%                      chirp mass in physical units)
  %%                      phi_star = phase at coalescence time [default=0]
  %% output parameters:   x = signal
  %%                      ifl = instantaneous frequency
  %%                      difl = 1st deriv. of the instantaneous frequency
  %%                      ddifl = 2nd deriv. of the instantaneous frequency
  %%                      chirp mass value if assuming fs=2kHz


  beta=5/8;
  
  if (nargin<1)
    error('usage: [x ifl]=gwchirp(N[,f_in,t_out,fs,phi_star])');
  end;
  
  if (nargin<2)
    f_in=.1;
  end;
  
  if (f_in<0)&(f_in>=.5)
    error('gwchirp: f_in should be >0 and <1/2');
  end;
  
  if (nargin<3)
    t_out=N;
  end;
  
  if ((t_out>N)|(t_out<=1))
    error('gwchirp:  t_out should be >1  and <=N');
  end;

  if (nargin<4)
    fs=2048; % Hz
  end;

  if (nargin<5)
    phi_star=0;
  end;

  %% -- compute mod. index and coalesc. time from inputs
  p=1/(beta-1);
  dbeta=((f_in^p-.5^p)/(t_out-1))^(beta-1);
  t_star=t_out+(.5/dbeta)^p;
  
  %% -- compute phase
  t=t_star-(1:t_out);
  phi=2*pi*dbeta/beta*t.^beta+phi_star;
  
  %% -- create and compute signal + ifl
  x=zeros(N,1);
  x(1:t_out)=exp(-i*phi);
  ifl=zeros(N,1);
  ifl(1:t_out)=dbeta*t.^(beta-1);
  difl=zeros(N,1);
  difl(1:t_out)=-dbeta*(beta-1)*t.^(beta-2);
  ddifl=zeros(N,1);
  ddifl(1:t_out)=dbeta*(beta-1)*(beta-2)*t.^(beta-3);
  
  %% -- invert dbeta to physical parameters
  ts=1/fs;
  c=160*3^(3/8)*beta*ts^beta;
  chirp_mass=(dbeta/c)^(-1/beta);

  %%printf('corr. chirp mass = %3.2f solar m.\n',chirp_mass);
  %%printf('assuming fs=%d Hz\n',fs);

endfunction;
