// sts.cc -- interface to tracksearch algorithm.

// To use this file, your version of Octave must support dynamic
// linking.  To find out if it does, type the command
//
//   x = octave_config_info; x.DEFS
//
// at the Octave prompt.  Support for dynamic linking is included if
// the output contains the string -DWITH_DYNAMIC_LINKING=1.
//
// You need to have the LAL (Ligo Algorithm Library, www.ligo.org) 
// to compile this code.
//
// Enter:
//
// help sts
//
// at the Octave prompt to have a description of how sts should be used.  
// 
// This file may be used under the terms defined by the GNU 
// General Public License [write to the Free Software Foundation, 
// 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA]
//
// WITHOUT ANY WARRANTY; without even the implied warranty 
// of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
//
// Copyright (C) 2005 Eric Chassande-Mottin, CNRS (France)
//

#include <octave/config.h>

#include <iostream>
#include <sstream>
using namespace std;
#include <string>

#include <octave/defun-dld.h>
#include <octave/error.h>
#include <octave/oct-obj.h>
#include <octave/pager.h>
#include <octave/symtab.h>
#include <octave/variables.h>

#include <lal/TrackSearch.h>
#include <lal/TimeFreq.h>

INT4 lalDebugLevel=0;

DEFUN_DLD (sts, args, ,
  "STS: track search in the time-frequency plane.\n\
Usage:\n\
[smax,N,f,t,idx] = sts(tfr[,opts])\n\
Input parameters:\n\
tfr : time-frequency representation\n\
opts : optional vector containing search parameters\n\
opts is a column vector of size 3x1\n\
opts(1) is the smoothing width of the Gaussian kernel used in\n\
the prefiltering of the TFR data (default=2)\n\
opts(2) is the high threshold (default=1)\n\
opts(3) is the low threshold (default=0.33)\n\
Output parameters:\n\
smax : maximum path integral along detected curves\n\
N : number of detected curves\n\
f,t : time-frequency coordinates of the curves\n\
idx : start index of each curves.")
{  
  octave_value_list retval;
  
  /*----------- get input parameters ----------------*/
  int nargin = args.length ();
  
  if (nargin < 1)
    {
      // print_usage("sts");
      error("Usage: [smax,n,x,y,idx] = sts(tfr[,opts])");
      return retval;
    }

  /* get the data */
  Matrix tfr=args(0).matrix_value();

  if (tfr.rows()<16)
    {
      // print_usage("sts");
      error("TFR must be larger than 16x16");
      return retval;
    }
  
  /* get thresholds and smoothing */
  double sigma=2.0;
  double thigh=1.0;
  double tlow=0.33;
  if (nargin > 1)
    {
      ColumnVector opts(args(1).vector_value());
      sigma=opts(0);
      thigh=opts(1);
      tlow=opts(2);
    }

  /* define struct */
  static LALStatus status;
  static TimeFreqRep in;
  static TrackSearchOut out;
  static TrackSearchParams params;
  
  /* set the parameters */
  params.sigma=sigma; 
  params.high=thigh;
  params.low=tlow;
  params.height=tfr.columns();
  params.width=tfr.rows();
  params.allocFlag=1;

  /* alloc memory */
  in.map = (REAL4**) malloc(params.height*sizeof(REAL4*));
  for(int i=0;i<params.height;i++)
    in.map[i] = (REAL4*) malloc(params.width*sizeof(REAL4));
  
  /* map the data to input struct */
  for(int j=0;j<params.width;j++)
    for(int i=0;i<params.height;i++)
      *(in.map[i] + j) = tfr(i,j);
  
  /* Search for curves */
  LALSignalTrackSearch(&status, &out, &in, &params);
  
  // REPORTSTATUS(&status);
  // printf("number of curves %d\n",out.numberOfCurves);

  int N=out.numberOfCurves;
  ColumnVector idx (N);
  int ltot=0;
  for(int i=0;i<N;i++)
    {
      idx(i)=ltot+1;
      ltot+=out.curves[i].n;
    }

  ColumnVector x (ltot);
  ColumnVector y (ltot);
  // ColumnVector s (N);
  double s, smax=0.0;
  for(int i=0;i<N;i++)
    {
      s=0.0;
      for(int j=0;j<out.curves[i].n;j++){
	x((int) idx(i)-1+j)=(double) out.curves[i].row[j]+1.0;
	y((int) idx(i)-1+j)=(double) out.curves[i].col[j]+1.0;	
	s+=tfr(out.curves[i].row[j],out.curves[i].col[j]);
      }
      if (s>smax)
	smax=s;
    }

  for(int i=0;i<N;i++){
    LALFree(out.curves[i].row);
    LALFree(out.curves[i].col);
  }
  if(out.curves!=NULL)
    LALFree(out.curves);
  out.curves=NULL;

  /* set params.allocFlag=2 and call the routine again to free space*/
  params.allocFlag=2;
  LALSignalTrackSearch(&status, &out, &in, &params);

  for(int i=0;i<params.height;i++)
    free(in.map[i]);
  free(in.map);

  retval(0)=smax;
  retval(1)=(double) N;
  retval(2)=x;
  retval(3)=y;
  retval(4)=idx;

  return retval;
}










